/* ====================================================================
 * Rpc_Gotcha_beta1.0
 * Chad Renfro <raw_sock@usa.net>
 * --------------------------------------
 *
 * Copyright (c) 1999 Raw_sock  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by
 *      raw_sock <raw_sock@usa.net>"
 *
 * 4. The names "Raw_sock" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    raw_sock@usa.net
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 */

#include <pcap/pcap.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <malloc.h>
#include <memory.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <rpc/rpc.h>
#include <rpc/pmap_prot.h>
#include <time.h>
#include "gotcha.h"


struct cnx {
        char s_ip[16];
        char d_ip[16];
        int  s_port;
        int  d_port;
        int  total_bytes;
        int  total_pkts;
	int  rpc_session;
        u_long xid;
        u_long program_num;
        time_t start_time;
        time_t last_pkt;
        struct cnx *next;
};
struct cnx *cnx_head, *cnx_tail;

/***********/
/* Globals */
/***********/
struct rpc_list{
        char   *name;
        u_long number;
        struct rpc_list *next;
};
struct rpc_list *rpc_list_head, *rpc_list_tail;

FILE *log_file, *rpc_list_file, *cnx_log_file, *rpc_alert_log_file;
static pcap_t *p_dev;
int optind, opterr, data_buff_size,VERBOSE=0,intranet_flag=0;
char *optarg, *rpc_data_buffer; 

/**************/
/* Prototypes */
/**************/
struct rpc_list *read_config(FILE *file, struct rpc_list *list_head, struct rpc_list *list_tail);
int fgetline(char **buf, int *size, FILE *fp);
void print_nodes(struct rpc_list *cnx_head);
void file_operation(void);
void pkt_snatcher(u_char *user_data, const struct pcap_pkthdr *pkt_hdr, const u_char *pkt);
int udp_print(register const u_char *udp_pkt, u_int length,
                register const u_char *ip_pkt);
int tcp_print(register const u_char *tcp_pkt, u_int length,
                register const u_char *ip_pkt);
void alert(struct cnx *node);
void add_state_node(char *sa,char *da, int sp, int dp,struct rpc_msg *rp);
int remove_state_node(char *sa, char *da,int sp,int dp);
void erase_old(void);
int update_state_node(char *sa, char *da,int sp,int dp,int total_bytes,struct rpc_msg *rp,char *data_ptr);
struct tm *get_time(void);
void log_call(struct cnx *node);
char *search_node(struct cnx *node, struct rpc_list *list_head);
void print_ascii(const char *start,const int len);
void print_hex(register const char *ptr, register u_int ptr_len);
void usage(void);



int main(int argc, char **argv) {
	register int  option, snaplen;
        char *device=NULL,error_buff[PCAP_ERRBUF_SIZE], *infile;
	u_int32_t localnet, netmask;

	snaplen = DEFAULT_SNAPLEN;
	opterr = 0;
	infile=NULL;
	while ((option = getopt(argc, argv, "hV:i:F:I")) != EOF)
		switch (option) {
		case 'V':
			++VERBOSE;
			break;
		case 'h':
			usage();
			break;
		case 'i':
			device = optarg;
			break;
		case 'F':
			infile = optarg;
			break;
		case 'I':
			++intranet_flag;
			break;
		default:
			break;
		}

        /* handle files */
	file_operation();
        if (infile != NULL) {
                p_dev = pcap_open_offline(infile, error_buff);
                if (p_dev == NULL)
                        error("%s", error_buff);
                localnet = 0;
                netmask = 0;
        } else {
	   if(!VERBOSE){
		if(fork() != 0)exit(0);
		fprintf(stderr,"\n\nProcess deamonized\n\n");
           }
		/* Determine the device name if not specified */
	  if (device == NULL && (device = pcap_lookupdev(error_buff)) == NULL) {
		perror(error_buff);
		exit(0);
          }
	  p_dev = pcap_open_live(device, snaplen, 1, 1000, error_buff);
	  if (p_dev == NULL) {
		perror(error_buff);
		fprintf(stderr,"Could not open packet capture device\n");
	  }
	  if (pcap_lookupnet(device, &localnet, &netmask, error_buff) < 0)
		perror(error_buff);
	  fprintf(stderr,"listening on %s\n",device);
	  fflush(stderr);
	}
	pcap_loop(p_dev, -1, pkt_snatcher, NULL);
	    perror("pcap_loop exited");
 return(0);
}

void file_operation(void) {
	if((cnx_log_file=fopen("/var/log/rpc_gotcha_log","a")) == NULL ) {
       		 perror("LOG_FILE");
       		 exit(1);
	}
	if((rpc_alert_log_file=fopen("/var/log/rpc_alert_log","a")) == NULL ) {
       		 perror("LOG_FILE");
       		 exit(1);
	}
	if((rpc_list_file=fopen("/etc/rpc_list.conf","r")) == NULL ) {
       		 perror("LOG_FILE");
       		 exit(1);
	}

	/* Now read files and put them into lists */
	rpc_list_head=read_config(rpc_list_file, rpc_list_head, rpc_list_tail);

}



struct rpc_list *read_config(FILE *file, struct rpc_list *list_head, struct rpc_list *list_tail) {
 int size=0;
 u_long tmp_num=0;
 struct rpc_list *new_node=NULL;
 char *f_buff=NULL;
  while (fgetline(&f_buff, &size, file) >= 0) {
        if(!(strstr(f_buff,"#"))) {
           new_node=(struct rpc_list *)malloc(sizeof(struct rpc_list));
	   new_node->name=strsep(&f_buff," \t");
	   tmp_num=atoi(strsep(&f_buff," \t"));
	   new_node->number=tmp_num;
	   new_node->next=NULL;
	   if(!list_head) { 
		list_head=new_node;
		list_tail=new_node;
	   } else {
		list_tail->next=new_node;
		list_tail=new_node;
		list_tail->next=NULL;
	   }
           if(VERBOSE)
		fprintf(stderr,"********** Node added to the tree *********\n");
        } /*if*/
  } /*while*/
 print_nodes(list_head);
 return(list_head);
}

void print_nodes(struct rpc_list *head) {
        struct rpc_list *current_node,*prev;

        if (head) {
            current_node=head;
            prev=head;
            while (current_node) {
		if(VERBOSE)
	           fprintf(stderr,"%ld\t\t%s\n",current_node->number,
                                                 current_node->name);
	      current_node=current_node->next;
            }/*while*/
        }/*if*/
}/*proc*/

int fgetline(char **buf, int *size, FILE *fp) {

  int len;
  int c;

  if (fp == NULL)
    return -1;
  if (*buf == NULL) {
    if (*size == 0) 
      *size = BUFSIZ;
    if ((*buf = malloc(*size)) == NULL)
      return -1;
  }
  if (feof(fp))
    return -1;
  len = 0;
  while (((c = getc(fp)) != EOF) && (c != '\n')) {
    if (len+1 >= *size) {
      if ((*buf = realloc(*buf, *size += BUFSIZ)) == NULL)
        return -1;
    }
    (*buf)[len++] = c;
  }
  if (len == 0 && c == EOF)
    return -1;
  (*buf)[len] = '\0';
  return len;
} /* fgetline */


void pkt_snatcher(u_char *user_data, const struct pcap_pkthdr *pkt_hdr, const u_char *pkt)
{
	register const struct ip *ip;
	register u_int hlen, len;
	register const u_char *trans_pkt;

	pkt += sizeof(struct ether_header);
	ip= (const struct ip *)pkt;
	hlen = ip->ip_hl * 4;		/* Total length of the ip header */
	len = ntohs(ip->ip_len); 	/* Total length of the ip datagram */
		trans_pkt = (const u_char *)ip + hlen;
		switch (ip->ip_p) {

		case IPPROTO_TCP:
			tcp_print(trans_pkt, len, (const u_char *)ip);
			break;

		case IPPROTO_UDP:
			udp_print(trans_pkt, len, (const u_char *)ip);
			break;
		default:
			break;
		}
}

int udp_print(register const u_char *udp_pkt, u_int length, register const u_char *ip_pkt) { 
	register const struct udphdr *udp;
	register struct cnx *new_cnx;
	register const struct ip *ip;
	register const u_char *cast_p;
	int sport, dport, ulen,tot_data_bytes;
        register struct rpc_msg *rp;
	enum msg_type direction;
	char *tmp_name=NULL;
	char sip[16],dip[16];
        u_long tmp_xid, tmp_num;

	udp = (struct udphdr *)udp_pkt;
	ip = (struct ip *)ip_pkt;
        strncpy(sip,(char *)inet_ntoa(ip->ip_src),16);
        strncpy(dip,(char *)inet_ntoa(ip->ip_dst),16);
	sport = ntohs(udp->source);
	dport = ntohs(udp->dest);
	if(dport < 110 || dport==443 || sport==20) {
		return(1);
	}
	cast_p = (u_char *)(udp + 1);
        ulen = ntohs(udp->len);
	tot_data_bytes = ulen - 8 - 24;

        /* The Real RPC Stuff */
	rp = (struct rpc_msg *)(udp + 1);
     	tmp_num=ntohl(rp->rm_call.cb_prog);
	direction = (enum msg_type)ntohl(rp->rm_direction);
	if ((direction == 0) && (ntohl(rp->rm_call.cb_rpcvers)==2)) {
           if(intranet_flag && tmp_num==100003)return(0);
       	   new_cnx=(struct cnx *)malloc(sizeof(struct cnx));
           strncpy(new_cnx->s_ip,sip,16);
           strncpy(new_cnx->d_ip,dip,16);
       	   new_cnx->total_pkts=1;
       	   new_cnx->total_bytes=tot_data_bytes;
     	   new_cnx->s_port=sport;
     	   new_cnx->d_port=dport;
     	   new_cnx->next=NULL;
     	   tmp_xid=ntohl(rp->rm_xid);
	   new_cnx->program_num=tmp_num;
	   log_call(new_cnx);
	   tmp_name = search_node(new_cnx,rpc_list_head);
	   if(VERBOSE)fprintf(stderr,"The program name is %s\n",tmp_name);
	   print_ascii(cast_p,ulen);
	   if(strstr(rpc_data_buffer,"bin/")) {
               alert(new_cnx);
	   }
           free(rpc_data_buffer);
	   free(new_cnx);
        } 
  return(0);
}

int tcp_print(register const u_char *tcp_pkt, u_int length, register const u_char *ip_pkt) {
        register const struct tcphdr *tp;
        register const struct ip *ip;
        register const u_char *cast_p;
        int sport, dport, hlen,data_len;
	char *data_ptr;
	char sip[15],dip[15];
        register struct rpc_msg *rp;


        tp = (struct tcphdr *)tcp_pkt;
        ip = (struct ip *)ip_pkt;
        sport = ntohs(tp->source);
        dport = ntohs(tp->dest);
	if(dport < 110 || dport==443 || sport==20){
		/* Dont evaluate WKS */
		return(1);
	}
	hlen = tp->doff * 4;
        cast_p = (u_char *)(tp + 1);
        length -= sizeof(struct tcphdr);
	data_len = ntohs(ip->ip_len) - ((ip->ip_hl*4) + (hlen));
	data_ptr = (char *)tcp_pkt + hlen;
        strncpy(sip,(char *)inet_ntoa(ip->ip_src),15);
        strncpy(dip,(char *)inet_ntoa(ip->ip_dst),15);

	
        rp = (struct rpc_msg *)(tcp_pkt + hlen + 4);
	if(tp->syn && !tp->ack) {
	  /* Start of a connection */
          add_state_node(sip,dip, sport, dport,rp);
	} else if(tp->rst || tp->fin) {
	  remove_state_node(sip,dip, sport, dport);
	  erase_old();
	} else {
	  update_state_node(sip,dip, sport, dport,data_len,rp,data_ptr);
	}/*state conditional*/
return(0);

}


void add_state_node(char *sa,char *da, int sp, int dp,struct rpc_msg *rp) {
        struct cnx *new_cnx;
	u_long tmp_xid;
	u_long tmp_direction;
	u_long tmp_num;

        new_cnx=(struct cnx *)malloc(sizeof(struct cnx));
	strncpy(new_cnx->s_ip,sa,16);
	strncpy(new_cnx->d_ip,da,16);
        new_cnx->s_port=sp;
        new_cnx->d_port=dp;
        time(&(new_cnx->start_time));
        time(&(new_cnx->last_pkt));
        new_cnx->total_bytes=0;
        new_cnx->total_pkts=1;
        new_cnx->next=NULL;
	tmp_direction=ntohl(rp->rm_direction);
	tmp_xid=ntohl(rp->rm_xid);
	tmp_num=ntohl(rp->rm_call.cb_prog);
	if((tmp_direction == 0) && (ntohl(rp->rm_call.cb_rpcvers)==2)) {
	  new_cnx->rpc_session=1;
	  new_cnx->xid = tmp_xid;
	  new_cnx->program_num=tmp_num;
	  log_call(new_cnx);
	  if(VERBOSE) {
	   fprintf(stderr,"The direction is ::: %d\n",ntohl(rp->rm_direction));
	   fprintf(stderr,"The rpc program_number is ::: %ld\n",tmp_num);
	  }
	} else {
	  new_cnx->rpc_session=0;
	}
        if (!cnx_head) {
          cnx_head=new_cnx;
          cnx_tail=new_cnx;
        } else {
          cnx_tail->next=new_cnx;
          cnx_tail=new_cnx;
          /* FORCE it */
          cnx_tail->next=NULL;
        }
} /*proc*/

int remove_state_node(char *sa, char *da,int sp,int dp) {
        struct cnx *current_node,*prev;
        int duration;
 
        if (cnx_head) {
          current_node=cnx_head;
          prev=cnx_head;
          while (current_node) {
           if ((((strcmp(sa,current_node->s_ip)==0) && 
	     (strcmp(da,current_node->d_ip)==0)) && 
	     ((sp==current_node->s_port)&&(dp==current_node->d_port)))) {
	      if(current_node->rpc_session) {
                duration=current_node->last_pkt - current_node->start_time;
	       }
                  prev->next=current_node->next;
                  if (current_node==cnx_head) {
                    cnx_head=cnx_head->next;
                    /* prev=NULL; */
                  }
                  if (current_node==cnx_tail) {
                    cnx_tail=prev;
                  }/*if*/
                  free (current_node);
                  return 1;
             }/*long if*/
             prev=current_node;
             current_node=current_node->next;
           }/*while*/
       }/*if*/
  return(0);
}/*proc*/

void erase_old(void) {
        struct cnx *current_cnx, *next;
        time_t present_time=0;

        current_cnx=cnx_head;
        time(&present_time);
        while(current_cnx!=NULL) {
          next = current_cnx->next;
          if(present_time - current_cnx->last_pkt > TIMEOUT) {
             remove_state_node(current_cnx->s_ip,current_cnx->d_ip,
             current_cnx->s_port,current_cnx->d_port); 
	  }
          current_cnx=next;
        }
}

int update_state_node(char *sa, char *da,int sp,int dp,int tot_data_bytes,struct rpc_msg *rp,char *data_ptr) {
        register struct cnx *current_node=NULL,*prev;
	static int update_cnt=0;

        if (cnx_head) {
            current_node=cnx_head;
            prev=cnx_head;
            while (current_node) {
             if ((((strcmp(sa,current_node->s_ip)==0) && 
		(strcmp(da,current_node->d_ip)==0)) && 
		((sp==current_node->s_port)&&(dp==current_node->d_port)))) {
                  time(&(current_node->last_pkt));
		  if(current_node->rpc_session) {
	             log_call(current_node);
		     print_ascii(data_ptr,tot_data_bytes);
		     if(strstr(rpc_data_buffer,"bin/")) {
			/* ALERT BUFFER OVERFLOW */
			alert(current_node);
		     }
                     free(rpc_data_buffer);
	        } else if((ntohl(rp->rm_direction) == 0) && 
                         (ntohl(rp->rm_call.cb_rpcvers)==2)) {
                     current_node->rpc_session=1;
                     current_node->xid = ntohl(rp->rm_xid);
                     current_node->program_num=ntohl(rp->rm_call.cb_prog);
		     log_call(current_node);
		     print_ascii(data_ptr,tot_data_bytes);
                     if(strstr(rpc_data_buffer,"bin/")) {
                        /* ALERT BUFFER OVERFLOW */
			alert(current_node);
                     }
                     free(rpc_data_buffer);
                }
                current_node->total_bytes=current_node->total_bytes + 
							tot_data_bytes;
                current_node->total_pkts++;
             }/*long if*/
              current_node=current_node->next;
           }/*while*/
       }/*if*/
  if(update_cnt++ > 500) {
	erase_old();
	update_cnt=0;
  }
  return(0);
}/*proc*/

struct tm *get_time(void) {
   time_t timenow;
        timenow = time(0);
        return(localtime(&timenow));
}

void alert(struct cnx *node) {
   struct tm *timestrptr;
   char currenttime[80];

   if(VERBOSE)fprintf(stderr,"IN THE ALERT FUNCTION\n\n\n");
   timestrptr = get_time();
   strftime(currenttime,80,"%c",timestrptr);	
    /* Write alert to log file */
   fprintf(rpc_alert_log_file,"\n******************************************\n");
   fprintf(rpc_alert_log_file,"Possible buffer overflow targeting %ld (%s)\n",
			    node->program_num,search_node(node,rpc_list_head));
   fprintf(rpc_alert_log_file,"Source ip address \t\t%s\n",node->s_ip);
   fprintf(rpc_alert_log_file,"Source port \t\t\t%d\n",node->s_port);
   fprintf(rpc_alert_log_file,"Dest ip address   \t\t%s\n",node->d_ip);
   fprintf(rpc_alert_log_file,"Dest port   \t\t\t%d\n",node->d_port);
   fprintf(rpc_alert_log_file,"Time of attack    \t\t%s\n",currenttime);
   fprintf(rpc_alert_log_file,"Data that was passed :\n");
   fprintf(rpc_alert_log_file,"---------------------\n%s\n",rpc_data_buffer);
   fprintf(rpc_alert_log_file,"\n******************************************\n");
   fflush(rpc_alert_log_file);
}
	
void log_call(struct cnx *node) {
 struct tm *timestrptr;
 char currenttime[80];

  timestrptr = get_time();
  strftime(currenttime,80,"%c",timestrptr);
  fprintf(cnx_log_file,"%s   %s:%d ---> %s:%d  %d  %ld (%s)\n",
        currenttime,node->s_ip,node->s_port,node->d_ip,
	node->d_port, node->total_pkts,
        node->program_num,
	search_node(node,rpc_list_head));
  fflush(cnx_log_file);
}


char *search_node(struct cnx *node, struct rpc_list *list_head) {
   struct rpc_list *current_node;

   if (list_head) {
      current_node=list_head;
      while (current_node) {
        if (node->program_num == current_node->number) {
	   return(current_node->name);
        }
        current_node=current_node->next;
      }/*while*/
   }/*if*/
  return(NULL);
}/*proc*/

void print_ascii(const char *data_ptr, const int len)
{
   int done=0;
   const char *data, *data_end;
   char *tmp_data_ptr;

   data_end = data_ptr + (len-1);
   rpc_data_buffer = (char *) malloc((len + 2) * sizeof(char));
   memset(rpc_data_buffer, 0x20, ((len + 2) * sizeof(char)));
   data = data_ptr ;
   tmp_data_ptr = rpc_data_buffer;
   while(1) {
       if(*data > 0x1F && *data < 0x7E)
          *tmp_data_ptr = (*data&0xFF);
       tmp_data_ptr++;
       if(data < data_end)
         data++;
       else {
	    if(VERBOSE)fprintf(stderr,"done printing ascii char \n");
            *tmp_data_ptr='\n';
            tmp_data_ptr++;
            *tmp_data_ptr='\n';
            tmp_data_ptr++;
            *tmp_data_ptr = 0;
            data_buff_size = (int) (tmp_data_ptr - rpc_data_buffer);
            if(VERBOSE)fwrite(rpc_data_buffer, data_buff_size, 1, stderr);
            return;
         }
   }
}

void usage(void) {

	fprintf(stderr,"USAGE :\n\t rpc_gotcha [-i interface] [-I IntrAnet use] [-F read file] \n");
	fprintf(stderr,"\t\t -I = internal network , no nfs traffic will be reported or logged\n");
	exit(0);
}

